#!/usr/bin/python3
"""
A quick hack to re-assemble XSDs distributed over many xs:includes.

Warning:

This doesn't do any namespace processing and assumes the XSD prefix is xs (you
can change that in XSD_PREFIX).  This will fail spectacularly if the prefix
mapping is different between parent and child documents or if child documents
add prefixes in their root elements.

Both would be fixable with reasonable effort.  See
https://blog.tfiu.de/inlining-xs-include-in-xml-schema.html
for comments and ways to encourage the author to improve this.

Distributed under CC-0 (a.k.a. public domain).
"""

import os
import sys
from xml.etree import ElementTree

from xml import sax
from xml.sax import saxutils
from xml.sax.handler import ContentHandler

XSD_PREFIX = "xs"
INCLUDE = f"{XSD_PREFIX}:include"
ROOT = f"{XSD_PREFIX}:schema"

class IncludeExpander(ContentHandler):
	def __init__(self, output, indent, xsd_path, strip_root):
		self.output, self.xsd_path = output, xsd_path
		self.strip_root = strip_root
		ContentHandler.__init__(self)
		self.indent = indent
		self.include_names = [None]

	def _formatAttributes(self, attrs):
		formatted = " ".join("{}={}".format(name, saxutils.quoteattr(value))
			for name, value in attrs.items())
		if formatted:
			return " "+formatted
		return ""

	def _emit(self, stuff):
		# we blindly encode everything as UTF-8, regardless of what the
		# input encoding was.  That's ok because without an XML declaration
		# (which we don't write), UTF-8 is implied.
		self.output.write(self.indent+stuff.encode("utf-8")+b"\n")

	def startElement(self, name, attrs):
		if name==ROOT and self.strip_root:
			# skip the root element if we're in the midst of another
			pass

		elif name==INCLUDE:
			included_path = os.path.join(
				self.xsd_path, 
				attrs["schemaLocation"])
			with open(included_path, "rb") as included:
				expand_file(
					included, 
					self.output, 
					self.indent, 
					os.path.dirname(included_path),
					True)

		else:
			self._emit("<{}{}>".format(
				name, self._formatAttributes(attrs)))
			self.indent += b"  "
	
	def endElement(self, name):
		if name==INCLUDE:
			pass

		elif name==ROOT and self.strip_root:
			pass

		else:
			self.indent = self.indent[:-2]
			self._emit(f"</{name}>")

	def characters(self, stuff):
		stuff = stuff.strip()
		if stuff:
			self._emit(stuff)


def expand_file(input, output, indent, xsd_path, strip_root=False):
	sax.parse(input, 
		IncludeExpander(output, indent, xsd_path, strip_root))


def main():
	with open(sys.argv[1], "rb") as input, open(sys.argv[2], "wb") as output:
		output.write(b"<!-- Stylesheet merged by expand-xsd-include.py -->\n")
		expand_file(input, output, b"", os.path.dirname(sys.argv[1]))


if __name__=="__main__":
	if len(sys.argv)!=3:
		sys.exit(f"Usage: {sys.argv[0]} <input-schema> <destination>")
	main()
